require 'wzcloud/transport'
require 'multi_json'
require 'ostruct'
require 'yaml'
require 'base64'

module WZCloud
  class Nova
    KEYPAIR_PREFIX = 'os-keypairs'.freeze
    ERROR_STATUS = 'ERROR'.freeze

    attr_reader :transport, :default_headers

    def initialize(access, transport, custom_headers = {})
      @access = access
      @transport = transport
      @base_url = @access.endpoints('nova')
      @transport.middleware.unshift(ErrorHandler)
      @default_headers = {'X-Auth-Token' => @access.token,
                          'Content-Type' => 'application/json',
                          'Accept' => 'application/json'}.merge(custom_headers)
    end

    def get_flavors(is_public = true)
      params = {}
      params[:is_public] = false unless is_public
      make_request(:get, 'flavors', nil, params)
        .body[:flavors].map(&OpenStruct.method(:new))
    end

    def get_flavor(id)
      make_request(:get, "flavors/#{id}").body[:flavor]
    end

    def get_flavors_detail(is_public = true)
      params = {}
      params[:is_public] = false unless is_public
      sort_flavors(make_request(:get, 'flavors/detail', nil, params).body[:flavors])
    end

    def get_flavor_extra_specs(flavor_id)
      make_request(:get, "flavors/#{flavor_id}/os-extra_specs", nil, {}).body[:extra_specs]
    end

    def get_flavor_access(flavor_id, options = {})
      make_request(:get, "flavors/#{flavor_id}/os-flavor-access", nil, options).body[:flavor_access]
    rescue WZCloud::Transport::NotFoundError => _e
      []
    end

    def get_images
      make_request(:get, 'images').body[:images].map(&OpenStruct.method(:new))
    end

    def get_images_detail(options = {})
      make_request(:get, 'images/detail', nil, options).body[:images]
    end

    def get_servers_detail(options = {})
      make_request(:get, 'servers/detail', nil, options).body[:servers]
    end

    def get_server(id)
      make_request(:get, "servers/#{id}").body[:server]
    end

    def get_servers(options = {})
      make_request(:get, 'servers', nil, options).body[:servers]
    end

    def create_server(name, flavor_ref, image_ref, network_id = nil,
                      av_zone = nil, key_name = nil, password = nil,
                      server_group = nil, count = 1, return_reservation_id = false)
      options = {server: {name: name, flavorRef: flavor_ref, imageRef: image_ref}}

      options['os:scheduler_hints'] = {group: server_group} if server_group
      options[:server][:availability_zone] = av_zone if av_zone
      options[:server][:key_name] = key_name if key_name
      options[:server][:networks] = Array(network_id).map { |nid| {uuid: nid} } if network_id
      options[:server][:min_count] = count if count.to_i > 1
      options[:server][:return_reservation_id] = 'True' if return_reservation_id
      unless password.to_s.empty?
        user_data = generate_user_data(password)
        options[:server][:user_data] = Base64.encode64(user_data)
      end

      result = make_request(:post, 'servers', options)

      return_reservation_id ? result.body : result.body[:server]
    end

    def rename_server(server_uuid, new_name)
      make_request(:put, "servers/#{server_uuid}", server: {name: new_name}).body[:server]
    end

    def rebuild_server(server_uuid, image_ref, name = nil)
      opts = {imageRef: image_ref}
      opts[:name] = name if name
      make_request(:post, "servers/#{server_uuid}/action", rebuild: opts).body[:server]
    end

    def delete_server(uuid)
      make_request(:delete, "servers/#{uuid}")
    end

    def resize_server(server_uuid, flavor_ref)
      server = get_server(server_uuid)
      raise InstanceError if server[:status] != 'ACTIVE'
      data = {resize: {flavorRef: flavor_ref}}
      make_request :post, "servers/#{server_uuid}/action", data
    end

    def confirm_resize_server(server_uuid)
      server = get_server(server_uuid)
      raise InstanceError if server[:status] != 'VERIFY_RESIZE'
      make_request :post, "servers/#{server_uuid}/action", confirmResize: nil
    end

    def revert_resize_server(server_uuid)
      server = get_server(server_uuid)
      raise InstanceError if server[:status] != 'VERIFY_RESIZE' # XXX Maybe delete all this checks?
      make_request :post, "servers/#{server_uuid}/action", revertResize: nil
    end

    def start_server(server_uuid)
      make_request :post, "servers/#{server_uuid}/action", 'os-start' => nil
    end

    def stop_server(server_uuid)
      make_request :post, "servers/#{server_uuid}/action", 'os-stop' => nil
    rescue Transport::ConflictError => e
      raise unless e.message['vm_state stopped']
    end

    def reboot_server(server_uuid, type = 'SOFT')
      data = {reboot: {type: type}}
      make_request :post, "servers/#{server_uuid}/action", data
    end

    def reset_server(server_uuid, type = 'error')
      data = {'os-resetState' => {state: type}}
      make_request :post, "servers/#{server_uuid}/action", data
    end

    def get_server_groups(options = {})
      make_request(:get, 'os-server-groups', options).body[:server_groups]
    end

    def create_server_group(name, policy)
      options = {name: name, policies: [policy]}
      make_request(:post, 'os-server-groups', server_group: options).body[:server_group]
    end

    def get_server_group(server_group_uuid)
      make_request(:get, "os-server-groups/#{server_group_uuid}").body[:server_group]
    end

    def delete_server_group(server_group_uuid)
      make_request(:delete, "os-server-groups/#{server_group_uuid}")
    end

    def vnc_url(server_uuid)
      data = {:'os-getVNCConsole' => {type: 'novnc'}}
      make_request(:post, "servers/#{server_uuid}/action", data).body
    end

    def get_actions(server_uuid)
      make_request(:get, "/servers/#{server_uuid}/os-instance-actions")
        .body[:instanceActions]
    end

    def get_action_detail(server_uuid, request_id)
      make_request(:get, "/servers/#{server_uuid}/os-instance-actions/#{request_id}")
        .body[:instanceAction]
    end

    def attach_interface(server_id, network_id)
      wait_for_servers

      make_request(
        :post,
        "servers/#{server_id}/os-interface",
        interfaceAttachment: {net_id: network_id}
      ).body[:interfaceAttachment]
    end

    def detach_interface(server_id, network_id, neutron)
      ifaces = make_request(:get, "servers/#{server_id}/os-interface").body[:interfaceAttachments]
      iface  = ifaces.find { |i| i[:net_id] == network_id }
      raise "Interface for network #{network_id} not found on #{server_id}" unless iface
      neutron.delete_port(iface[:port_id])
    end

    def get_keypairs
      make_request(:get, KEYPAIR_PREFIX).body[:keypairs]
    end

    def get_keypair_by_fingerprint(fp)
      get_keypairs.find { |kp| kp[:keypair][:fingerprint] == fp } ||
        raise(ArgumentError, "Keypair with fingerprint #{fp} not found")
    end

    def create_keypair(name, public_key = nil)
      body = {keypair: {name: name}}
      body[:keypair][:public_key] = public_key if public_key
      begin
        make_request(:post, KEYPAIR_PREFIX, body).body[:keypair]
      rescue Transport::BadRequestError => e
        # Can't reproduce message
        if e.message['malformed or otherwise incorrect'] || e.message['Keypair data is invalid:']
          raise ArgumentError, 'Invalid public key'
        end
        raise e
      end
    end

    def delete_keypair(name)
      make_request :delete, "#{KEYPAIR_PREFIX}/#{name}"
    end

    def get_hosts
      hosts = make_request(:get, 'os-hosts').body[:hosts]
      # Filtering via query parameter doesn't work.
      hosts.select { |host| host[:service] == 'compute' }
    end

    def get_av_zones
      info = make_request(:get, 'os-availability-zone').body[:availabilityZoneInfo]
      info.delete_if { |zone| !zone[:zoneState][:available] }.map { |zone| zone[:zoneName] }
    end

    def get_snapshots(instance_id = nil)
      params = if instance_id
                 {server: "#{@base_url}/servers/#{instance_id}"}
               else
                 {}
               end
      images = get_images_detail(params)
      if instance_id
        images
      else
        images.select { |i| i[:server] }
      end
    end

    def create_snapshot(instance_id, options = {})
      name = options[:name] || "#{instance_id}-snap-#{Time.new.to_i}"
      metadata = options[:metadata] || {}
      begin
        r = make_request(
          :post,
          "servers/#{instance_id}/action",
          createImage: {name: name, metadata: metadata}
        )
        {id: r.headers['Location'][%r{\/([-a-f0-9]+)\z}, 1]}
      rescue Transport::RequestError => e
        case e
        when Transport::NotFoundError
          raise ArgumentError, "Instance #{instance_id} not found"
        when Transport::ConflictError
          raise ArgumentError, "Instance #{instance_id} is in conflicting state"
        else
          raise e
        end
      end
    end

    def delete_image(id)
      make_request(:delete, "images/#{id}")
    end

    def fetch_local_ip(server_details)
      local_ip = nil
      server_details[:addresses].each do |addr_name, ips|
        next unless addr_name['local-network']
        ips.each do |ip|
          local_ip = ip[:addr] if ip[:"OS-EXT-IPS:type"] == 'fixed'
        end
      end
      local_ip
    end

    def fetch_internal_and_external_ips(server_details)
      internal_ip = nil
      external_ip = nil
      server_details[:addresses].each do |addr_name, ips|
        next unless addr_name['network-for']
        ips.each do |ip|
          case ip[:"OS-EXT-IPS:type"]
          when 'fixed' then internal_ip = ip[:addr]
          when 'floating' then external_ip = ip[:addr]
          end
        end
      end
      {internal_ip: internal_ip, external_ip: external_ip}
    end

    def quotas(tenant_id)
      make_request(:get, "os-quota-sets/#{tenant_id}").body[:quota_set]
    end

    def update_quotas(tenant_id, new_quotas)
      data = {quota_set: new_quotas}
      make_request(:put, "os-quota-sets/#{tenant_id}", data).body[:quota_set]
    end

    # Requires admin privileges by default, but that can be changed via
    # +compute_extension:os-server-external-events:create+ policy.
    def refresh_network(server_id)
      data = {events: [{name: 'network-changed', server_uuid: server_id}]}
      make_request(:post, 'os-server-external-events', data)
    rescue WZCloud::Transport::NotFoundError
      # Network not available yet? Nothing to reset, then.
    end

    def get_console_output(instance_id)
      make_request(:post, "servers/#{instance_id}/action",
                   :"os-getConsoleOutput" => {length: nil})
        .body[:output]
    end

    def migrate_server(instance_id)
      make_request(:post, "servers/#{instance_id}/action", migrate: nil)
    end

    def get_volumes(server_id)
      make_request(:get, "servers/#{server_id}/os-volume_attachments").body[:volumeAttachments]
    end

    def attach_volume(server_id, volume_id)
      data = {volumeAttachment: {volumeId: volume_id}}
      make_request(
        :post,
        "servers/#{server_id}/os-volume_attachments",
        data
      ).body[:volumeAttachment]
    end

    def rescue_server(server_uuid, admin_password = nil, rescue_image_ref = nil)
      data = {rescue: {}}
      data[:rescue][:adminPass] = admin_password if admin_password
      data[:rescue][:rescue_image_ref] = rescue_image_ref if rescue_image_ref
      make_request(:post, "servers/#{server_uuid}/action", data).body
    end

    def unrescue_server(server_uuid)
      data = {unrescue: nil}
      make_request(:post, "servers/#{server_uuid}/action", data).body
    end

    def detach_volume(server_id, attachment_id)
      make_request(:delete, "servers/#{server_id}/os-volume_attachments/#{attachment_id}")
    end

    def lock(server_uuid)
      make_request :post, "servers/#{server_uuid}/action", lock: nil
    end

    def unlock(server_uuid)
      make_request :post, "servers/#{server_uuid}/action", unlock: nil
    end

    def show_attachment(server_id, attachment_id)
      make_request(
        :get,
        "servers/#{server_id}/os-volume_attachments/#{attachment_id}"
      ).body[:volumeAttachment]
    end

    def wait_for_servers
      loop do
        servers = get_servers_detail
        break if !servers.any? { |serv| %w(BUILD RESIZE).include? serv[:status] } &&
                 !servers.any? { |serv| ['deleting'].include? serv[:"OS-EXT-STS:task_state"] }
        sleep 1
      end
    end

    def wait_status(server, status, seconds = 120)
      server_status = get_server(server[:id])[:status]
      seconds.times do
        server_status = get_server(server[:id])[:status]
        break if server_status == status || server_status == ERROR_STATUS
        sleep 1
      end
      raise InstanceError, server_status if server_status != status
    end

    def wait_for_snapshots
      loop do
        images = get_images_detail
        break unless images.any? { |image| ['SAVING'].include? image[:status] }
        sleep 1
      end
    end

    def clear_db(servers = get_servers)
      servers.each do |server|
        begin
          delete_server(server[:id])
        rescue Transport::RequestError => e
          # that happens sometimes
          raise e unless e.status == 404
        end
      end

      get_snapshots.each do |snap|
        begin
          delete_image(snap[:id])
        rescue Transport::RequestError => e
          # that happens too
          raise e unless e.status == 404
        end
      end

      # If we don't wait until all instances are terminated,
      # the next cleanup action is likely to remove the current tenant,
      # and then the remaining instances won't get deleted automatically.
      while servers.any?
        servers.each do |server|
          begin
            get_server(server[:id])
          rescue Transport::NotFoundError
            servers.delete(server)
          end
        end

        sleep 0.1
      end

      get_keypairs.each do |keypair|
        delete_keypair keypair[:keypair][:name]
      end
    end

    def make_request(method, url_path, body = nil, params = nil)
      json_body = MultiJson.dump(body) if body

      @transport.http_call("#{@base_url}/#{url_path}", method,
                           default_headers, json_body, params)
    end

    private

    def generate_user_data(password)
      %(
#cloud-config
disable_root: False
ssh_pwauth: True
users: root
chpasswd:
  list: |
    root:#{password}
  expire: True
      ).strip
    end

    def sort_flavors(flavors)
      flavors.sort { |f1, f2| f1[:disk] <=> f2[:disk] }
    end

    class InstanceError < StandardError; end

    class ErrorHandler
      def initialize(app)
        @app = app
      end

      def call(envir)
        @app.call(envir)
      rescue Faraday::Error::TimeoutError => e
        raise Transport::TimeoutError.new(envir[:url].to_s, 'Timeout error', 504)
      rescue Faraday::Error::ClientError => e
        raise unless e.response
        case e.response[:status]
        when 400
          message = e.response[:body][:badRequest][:message]
          raise Transport::BadRequestError.new(envir[:url].to_s, message, 400)
        when 403
          raise Transport::UnauthorizedError.new(
            envir[:url].to_s,
            e.response[:body].values.first[:message],
            403
          )
        when 404
          message = if e.response[:body].class == String
                      e.response[:body]
                    else
                      e.response[:body][:itemNotFound][:message]
                    end
          raise Transport::NotFoundError.new(envir[:url].to_s, message, 404)
        when 409
          message = e.response[:body][:conflictingRequest][:message]
          raise Transport::ConflictError.new(envir[:url].to_s, message, 409)
        else
          body = e.response[:body]
          message = body.values.first[:message] if body && body.class == Hash
          raise Transport::RequestError.new(envir[:url].to_s, message, e.response[:status])
        end
      end
    end
  end
end
